@@ -157,12 +157,23 @@ void export_config(py::module_& m) {
         .def("concurrent_lookup_requests", &ClientConfiguration::getConcurrentLookupRequest)
         .def("concurrent_lookup_requests", &ClientConfiguration::setConcurrentLookupRequest,
              return_value_policy::reference)
+        .def("stats_interval_in_seconds", &ClientConfiguration::getStatsIntervalInSeconds)
+        .def("stats_interval_in_seconds", &ClientConfiguration::setStatsIntervalInSeconds,
+             return_value_policy::reference)
         .def("use_tls", &ClientConfiguration::isUseTls)
         .def("use_tls", &ClientConfiguration::setUseTls, return_value_policy::reference)
         .def("tls_trust_certs_file_path", &ClientConfiguration::getTlsTrustCertsFilePath,
              return_value_policy::copy)
         .def("tls_trust_certs_file_path", &ClientConfiguration::setTlsTrustCertsFilePath,
              return_value_policy::reference)
+        .def("tls_private_key_file_path", &ClientConfiguration::getTlsPrivateKeyFilePath,
+             return_value_policy::copy)
+        .def("tls_private_key_file_path", &ClientConfiguration::setTlsPrivateKeyFilePath,
+             return_value_policy::reference)
+        .def("tls_certificate_file_path", &ClientConfiguration::getTlsCertificateFilePath,
+             return_value_policy::copy)
+        .def("tls_certificate_file_path", &ClientConfiguration::setTlsCertificateFilePath,
+             return_value_policy::reference)
         .def("tls_allow_insecure_connection", &ClientConfiguration::isTlsAllowInsecureConnection)
         .def("tls_allow_insecure_connection", &ClientConfiguration::setTlsAllowInsecureConnection,
              return_value_policy::reference)
@@ -302,7 +313,11 @@ void export_config(py::module_& m) {
         .def("batch_index_ack_enabled", &ConsumerConfiguration::setBatchIndexAckEnabled,
              return_value_policy::reference)
         .def("dead_letter_policy", &ConsumerConfiguration::setDeadLetterPolicy)
-        .def("dead_letter_policy", &ConsumerConfiguration::getDeadLetterPolicy, return_value_policy::copy);
+        .def("dead_letter_policy", &ConsumerConfiguration::getDeadLetterPolicy, return_value_policy::copy)
+        .def("crypto_failure_action", &ConsumerConfiguration::getCryptoFailureAction,
+            return_value_policy::copy)
+        .def("crypto_failure_action", &ConsumerConfiguration::setCryptoFailureAction,
+            return_value_policy::reference);
 
     class_<ReaderConfiguration, std::shared_ptr<ReaderConfiguration>>(m, "ReaderConfiguration")
         .def(init<>())
@@ -320,5 +335,9 @@ void export_config(py::module_& m) {
         .def("read_compacted", &ReaderConfiguration::setReadCompacted)
         .def("crypto_key_reader", &ReaderConfiguration::setCryptoKeyReader, return_value_policy::reference)
         .def("start_message_id_inclusive", &ReaderConfiguration::isStartMessageIdInclusive)
-        .def("start_message_id_inclusive", &ReaderConfiguration::setStartMessageIdInclusive, return_value_policy::reference);
+        .def("start_message_id_inclusive", &ReaderConfiguration::setStartMessageIdInclusive, return_value_policy::reference)
+        .def("crypto_failure_action", &ReaderConfiguration::getCryptoFailureAction,
+            return_value_policy::copy)
+        .def("crypto_failure_action", &ReaderConfiguration::setCryptoFailureAction,
+            return_value_policy::reference);
 }
