@@ -22,6 +22,7 @@
 import random
 import threading
 import logging
+from typing import Optional
 from unittest import TestCase, main
 import time
 import os
@@ -51,8 +52,6 @@ from pulsar.schema import JsonSchema, Record, Integer
 
 from _pulsar import ProducerConfiguration, ConsumerConfiguration, RegexSubscriptionMode
 
-from schema_test import *
-
 from urllib.request import urlopen, Request
 
 TM = 10000  # Do not wait forever in tests
@@ -483,6 +482,63 @@ class PulsarTest(TestCase):
 
         client.close()
 
+    def test_encryption_failure(self):
+        publicKeyPath = CERTS_DIR + "public-key.client-rsa.pem"
+        privateKeyPath = CERTS_DIR + "private-key.client-rsa.pem"
+        crypto_key_reader = CryptoKeyReader(publicKeyPath, privateKeyPath)
+        client = Client(self.serviceUrl)
+        topic = "my-python-test-end-to-end-encryption-failure-" + str(time.time())
+        producer = client.create_producer(
+            topic=topic, encryption_key="client-rsa.pem", crypto_key_reader=crypto_key_reader
+        )
+        producer.send(b"msg-0")
+
+        def verify_next_message(value: bytes):
+            consumer = client.subscribe(topic, subscription,
+                                        crypto_key_reader=crypto_key_reader)
+            msg = consumer.receive(3000)
+            self.assertEqual(msg.data(), value)
+            consumer.acknowledge(msg)
+            consumer.close()
+
+        subscription = "my-sub"
+        consumer = client.subscribe(topic, subscription,
+                                    initial_position=InitialPosition.Earliest,
+                                    crypto_failure_action=pulsar.ConsumerCryptoFailureAction.FAIL)
+        with self.assertRaises(pulsar.Timeout):
+            consumer.receive(3000)
+        consumer.close()
+        producer.send(b"msg-1")
+        verify_next_message(b"msg-0") # msg-0 won't be skipped
+
+        consumer = client.subscribe(topic, subscription,
+                                    initial_position=InitialPosition.Earliest,
+                                    crypto_failure_action=pulsar.ConsumerCryptoFailureAction.DISCARD)
+        with self.assertRaises(pulsar.Timeout):
+            consumer.receive(3000)
+        consumer.close()
+
+        producer.send(b"msg-2")
+        verify_next_message(b"msg-2") # msg-1 is skipped since the crypto failure action is DISCARD
+
+        # Encrypted messages will be consumed since the crypto failure action is CONSUME
+        consumer = client.subscribe(topic, 'another-sub',
+                                    initial_position=InitialPosition.Earliest,
+                                    crypto_failure_action=pulsar.ConsumerCryptoFailureAction.CONSUME)
+        for i in range(3):
+            msg = consumer.receive(3000)
+            self.assertNotEqual(msg.data(), f"msg-{i}".encode())
+            self.assertTrue(len(msg.data()) > 5, f"msg.data() is {msg.data()}")
+
+        reader = client.create_reader(topic, MessageId.earliest,
+                                      crypto_failure_action=pulsar.ConsumerCryptoFailureAction.CONSUME)
+        for i in range(3):
+            msg = reader.read_next(3000)
+            self.assertNotEqual(msg.data(), f"msg-{i}".encode())
+            self.assertTrue(len(msg.data()) > 5, f"msg.data() is {msg.data()}")
+
+        client.close()
+
     def test_tls_auth3(self):
         authPlugin = "tls"
         authParams = "tlsCertFile:%s/client-cert.pem,tlsKeyFile:%s/client-key.pem" % (CERTS_DIR, CERTS_DIR)
@@ -830,6 +886,7 @@ class PulsarTest(TestCase):
         self._check_value_error(lambda: Client(None))
         self._check_value_error(lambda: Client(self.serviceUrl, authentication="test"))
         self._check_value_error(lambda: Client(self.serviceUrl, operation_timeout_seconds="test"))
+        self._check_value_error(lambda: Client(self.serviceUrl, stats_interval_in_seconds="test"))
         self._check_value_error(lambda: Client(self.serviceUrl, io_threads="test"))
         self._check_value_error(lambda: Client(self.serviceUrl, message_listener_threads="test"))
         self._check_value_error(lambda: Client(self.serviceUrl, concurrent_lookup_requests="test"))
@@ -1021,11 +1078,22 @@ class PulsarTest(TestCase):
         msg = consumer.receive(TM)
         self.assertEqual(msg.data(), b"hello-0")
 
+        # seek with wrong type
+        with self.assertRaises(ValueError, msg="invalid seek_arg type <class 'float'>"):
+            consumer.seek(1.0)
+
         # seek on messageId
         consumer.seek(ids[50])
         msg = consumer.receive(TM)
         self.assertEqual(msg.data(), b"hello-51")
 
+        # seek on a user provided MessageId
+        msg_id = MessageId(ledger_id=ids[60].ledger_id(),
+                           entry_id=ids[60].entry_id())
+        consumer.seek(msg_id)
+        msg = consumer.receive(TM)
+        self.assertEqual(msg.data(), b"hello-61")
+
         # ditto, but seek on timestamp
         consumer.seek(timestamps[42])
         msg = consumer.receive(TM)
@@ -1036,6 +1104,10 @@ class PulsarTest(TestCase):
         with self.assertRaises(pulsar.Timeout):
             reader.read_next(100)
 
+        # seek with wrong type
+        with self.assertRaises(ValueError, msg="invalid seek_arg type <class 'float'>"):
+            consumer.seek(1.0)
+
         # earliest
         reader.seek(MessageId.earliest)
         msg = reader.read_next(TM)
@@ -1050,6 +1122,13 @@ class PulsarTest(TestCase):
         msg = reader.read_next(TM)
         self.assertEqual(msg.data(), b"hello-35")
 
+        # seek on a user provided MessageId
+        msg_id = MessageId(ledger_id=ids[44].ledger_id(),
+                           entry_id=ids[44].entry_id())
+        reader.seek(msg_id)
+        msg = reader.read_next(TM)
+        self.assertEqual(msg.data(), b"hello-45")
+
         # seek on timestamp
         reader.seek(timestamps[79])
         msg = reader.read_next(TM)
@@ -1231,6 +1310,29 @@ class PulsarTest(TestCase):
         s = MessageId.latest.serialize()
         self.assertEqual(MessageId.deserialize(s), MessageId.latest)
 
+        client = Client(self.serviceUrl)
+        topic = f'test-message-id-compare-{str(time.time())}'
+        producer = client.create_producer(topic)
+        consumer = client.subscribe(topic, 'sub')
+
+        sent_ids = []
+        received_ids = []
+        for i in range(5):
+            sent_ids.append(MessageId.wrap(producer.send(b'msg-%d' % i)))
+            msg = consumer.receive(TM)
+            received_ids.append(MessageId.wrap(msg.message_id()))
+            self.assertEqual(sent_ids[i], received_ids[i])
+            consumer.acknowledge(received_ids[i])
+        consumer.acknowledge_cumulative(received_ids[4])
+
+        for i in range(4):
+            self.assertLess(sent_ids[i], sent_ids[i + 1])
+            self.assertLessEqual(sent_ids[i], sent_ids[i + 1])
+            self.assertGreater(sent_ids[i + 1], sent_ids[i])
+            self.assertGreaterEqual(sent_ids[i + 1], sent_ids[i])
+            self.assertNotEqual(sent_ids[i], sent_ids[i + 1])
+        client.close()
+
     def test_get_topics_partitions(self):
         client = Client(self.serviceUrl)
         topic_partitioned = "persistent://public/default/test_get_topics_partitions"
@@ -1426,29 +1528,6 @@ class PulsarTest(TestCase):
         second_encode = schema.encode(record)
         self.assertEqual(first_encode, second_encode)
 
-    def test_configure_log_level(self):
-        client = pulsar.Client(
-            service_url="pulsar://localhost:6650",
-            logger=pulsar.ConsoleLogger(pulsar.LoggerLevel.Debug)
-        )
-
-        producer = client.create_producer(
-            topic='test_log_level'
-        )
-
-        producer.send(b'hello')
-
-    def test_configure_log_to_file(self):
-        client = pulsar.Client(
-            service_url="pulsar://localhost:6650",
-            logger=pulsar.FileLogger(pulsar.LoggerLevel.Debug, 'test.log')
-        )
-
-        producer = client.create_producer(
-            topic='test_log_to_file'
-        )
-
-        producer.send(b'hello')
 
     def test_logger_thread_leaks(self):
         def _do_connect(close):
@@ -1513,10 +1592,22 @@ class PulsarTest(TestCase):
         with self.assertRaises(TypeError):
             fun()
 
-    def _test_basic_auth(self, id, auth):
-        client = Client(self.adminUrl, authentication=auth)
+    def _test_basic_auth(self, topic_id: int, auth,
+                         use_tls: bool = False,
+                         tls_private_key_file_path: Optional[str] = None,
+                         tls_certificate_file_path: Optional[str] = None) -> None:
+        if use_tls:
+            service_url = self.serviceUrlTls
+            tls_trust_certs_file_path = CERTS_DIR + 'cacert.pem'
+        else:
+            service_url = self.adminUrl
+            tls_trust_certs_file_path = None
+        client = Client(service_url, authentication=auth,
+                        tls_trust_certs_file_path=tls_trust_certs_file_path,
+                        tls_private_key_file_path=tls_private_key_file_path,
+                        tls_certificate_file_path=tls_certificate_file_path)
 
-        topic = "persistent://private/auth/my-python-topic-basic-auth-" + str(id)
+        topic = "persistent://private/auth/my-python-topic-basic-auth-" + str(topic_id)
         consumer = client.subscribe(topic, "my-sub", consumer_type=ConsumerType.Shared)
         producer = client.create_producer(topic)
         producer.send(b"hello")
@@ -1548,6 +1639,17 @@ class PulsarTest(TestCase):
                 auth_params_string='{{"username": "{}","password": "{}", "method": "unknown"}}'.format(username, password)
             ))
 
+    def test_tls_encryption_with_other_auth(self):
+        self._test_basic_auth(6, AuthenticationBasic('admin', '123456'),
+                              use_tls=True,
+                              tls_private_key_file_path=CERTS_DIR + 'client-key.pem',
+                              tls_certificate_file_path=CERTS_DIR + 'client-cert.pem')
+        with self.assertRaises(pulsar.ConnectError):
+            self._test_basic_auth(7, AuthenticationBasic('admin', '123456'),
+                                  use_tls=True,
+                                  tls_private_key_file_path=CERTS_DIR + 'client-cert.pem',
+                                  tls_certificate_file_path=CERTS_DIR + 'client-key.pem')
+
     def test_invalid_basic_auth(self):
         username = "invalid"
         password = "123456"
@@ -1885,5 +1987,13 @@ class PulsarTest(TestCase):
 
         client.close()
 
+    def test_consumer_name(self):
+        client = Client(self.serviceUrl)
+        name = 'my-consumer-name'
+        consumer = client.subscribe('test_consumer_name', 'sub', consumer_name=name)
+        self.assertEqual(consumer.consumer_name(), name)
+        client.close()
+
+
 if __name__ == "__main__":
     main()
